@@ -7,17 +7,27 @@ from trytond.report import Report
 from trytond.pyson import Eval, If
 from trytond.transaction import Transaction
 from trytond.pool import Pool, PoolMeta
+from trytond.backend.database import CursorInterface
+
+try:
+    import pytz
+    TIMEZONES = [(x, x) for x in pytz.common_timezones]
+except ImportError:
+    TIMEZONES = []
+TIMEZONES += [(None, '')]
+
+CursorInterface.cache_keys.update({'company', 'employee'})
 
 __all__ = ['Company', 'Employee', 'UserEmployee', 'User', 'Property',
-    'Sequence', 'SequenceStrict', 'CompanyConfigStart', 'CompanyConfig',
-    'CompanyReport', 'LetterReport']
+    'Sequence', 'SequenceStrict', 'Date', 'CompanyConfigStart',
+    'CompanyConfig', 'CompanyReport', 'LetterReport']
 __metaclass__ = PoolMeta
 
 
 class Company(ModelSQL, ModelView):
     'Company'
     __name__ = 'company.company'
-    _inherits = {'party.party': 'party'}
+    _rec_name = 'party'
     party = fields.Many2One('party.party', 'Party', required=True,
             ondelete='CASCADE')
     parent = fields.Many2One('company.company', 'Parent')
@@ -25,36 +35,24 @@ class Company(ModelSQL, ModelView):
     header = fields.Text('Header')
     footer = fields.Text('Footer')
     currency = fields.Many2One('currency.currency', 'Currency', required=True)
+    timezone = fields.Selection(TIMEZONES, 'Timezone', translate=False)
     employees = fields.One2Many('company.employee', 'company', 'Employees')
 
     @classmethod
     def __setup__(cls):
         super(Company, cls).__setup__()
-        cls._constraints += [
-            ('check_recursion', 'recursive_companies'),
-            ]
-        cls._error_messages.update({
-                'recursive_companies': \
-                    'You can not create recursive companies!',
-                })
 
     @classmethod
-    def copy(cls, companies, default=None):
-        Party = Pool().get('party.party')
+    def validate(cls, companies):
+        super(Company, cls).validate(companies)
+        cls.check_recursion(companies)
 
-        if default is None:
-            default = {}
-        default = default.copy()
-        new_companies = []
-        for company in companies:
-            default['party'], = Party.copy([company.party])
-            new_company, = super(Company, cls).copy([company], default=default)
-            new_companies.append(new_company)
-        return new_companies
+    def get_rec_name(self, name):
+        return self.party.rec_name
 
     @classmethod
-    def write(cls, companies, vals):
-        super(Company, cls).write(companies, vals)
+    def write(cls, companies, values, *args):
+        super(Company, cls).write(companies, values, *args)
         # Restart the cache on the domain_get method
         Pool().get('ir.rule')._domain_get_cache.clear()
 
@@ -62,10 +60,17 @@ class Company(ModelSQL, ModelView):
 class Employee(ModelSQL, ModelView):
     'Employee'
     __name__ = 'company.employee'
-    _inherits = {'party.party': 'party'}
+    _rec_name = 'party'
     party = fields.Many2One('party.party', 'Party', required=True)
     company = fields.Many2One('company.company', 'Company', required=True)
 
+    @staticmethod
+    def default_company():
+        return Transaction().context.get('company')
+
+    def get_rec_name(self, name):
+        return self.party.rec_name
+
 
 class UserEmployee(ModelSQL):
     'User - Employee'
@@ -78,18 +83,17 @@ class UserEmployee(ModelSQL):
 
 class User:
     __name__ = 'res.user'
-    main_company = fields.Many2One('company.company', 'Main Company',
-        on_change=['main_company'])
+    main_company = fields.Many2One('company.company', 'Main Company')
     company = fields.Many2One('company.company', 'Current Company',
         domain=[('parent', 'child_of', [Eval('main_company')], 'parent')],
-        depends=['main_company'], on_change=['company', 'employees'])
+        depends=['main_company'])
     companies = fields.Function(fields.One2Many('company.company', None,
         'Current Companies'), 'get_companies')
     employees = fields.Many2Many('res.user-company.employee', 'user',
         'employee', 'Employees')
     employee = fields.Many2One('company.employee', 'Current Employee',
         domain=[
-            ('company', '=', Eval('company')),
+            ('company', '=', Eval('company', -1)),
             ('id', 'in', Eval('employees', [])),
             ],
         depends=['company', 'employees'])
@@ -135,16 +139,18 @@ class User:
     def get_status_bar(self, name):
         status = super(User, self).get_status_bar(name)
         if self.company:
-            status += ' - %s [%s]' % (self.company.name,
+            status += ' - %s [%s]' % (self.company.rec_name,
                 self.company.currency.name)
         return status
 
+    @fields.depends('main_company')
     def on_change_main_company(self):
         return {
             'company': self.main_company.id if self.main_company else None,
             'employee': None,
             }
 
+    @fields.depends('company', 'employees')
     def on_change_company(self):
         Employee = Pool().get('company.employee')
         result = {
@@ -170,10 +176,11 @@ class User:
                 res['main_company.rec_name'] = user.main_company.rec_name
             res['employees'] = [e.id for e in user.employees]
         if user.employee:
-            res['employee'] = None
-            if user.employee:
-                res['employee'] = user.employee.id
-                res['employee.rec_name'] = user.employee.rec_name
+            res['employee'] = user.employee.id
+            res['employee.rec_name'] = user.employee.rec_name
+        if user.company:
+            res['company'] = user.company.id
+            res['company.rec_name'] = user.company.rec_name
         return res
 
     @classmethod
@@ -193,6 +200,7 @@ class User:
 
         if 'company' in res['fields']:
             selection = convert2selection(res['fields'], 'company')
+            selection.append((None, ''))
             user = cls(Transaction().user)
             if user.main_company:
                 companies = Company.search([
@@ -211,8 +219,10 @@ class User:
             user_id = Transaction().context['user']
         result = super(User, cls).read(ids, fields_names=fields_names)
         if (fields_names
-                and 'company' in fields_names
-                and 'company' in Transaction().context):
+                and (('company' in fields_names
+                        and 'company' in Transaction().context)
+                    or ('employee' in fields_names
+                        and 'employee' in Transaction().context))):
             values = None
             if int(user_id) in ids:
                 for vals in result:
@@ -220,17 +230,28 @@ class User:
                         values = vals
                         break
             if values:
-                main_company_id = values.get('main_company')
-                if not main_company_id:
-                    main_company_id = cls.read([user_id],
-                        ['main_company'])[0]['main_company']
-                companies = Company.search([
-                        ('parent', 'child_of', [main_company_id]),
-                        ])
-                company_id = Transaction().context['company']
-                if ((company_id and company_id in map(int, companies))
-                        or not company_id):
-                    values['company'] = company_id
+                if ('company' in fields_names
+                        and 'company' in Transaction().context):
+                    main_company_id = values.get('main_company')
+                    if not main_company_id:
+                        main_company_id = cls.read([user_id],
+                            ['main_company'])[0]['main_company']
+                    companies = Company.search([
+                            ('parent', 'child_of', [main_company_id]),
+                            ])
+                    company_id = Transaction().context['company']
+                    if ((company_id and company_id in map(int, companies))
+                            or not company_id):
+                        values['company'] = company_id
+                if ('employee' in fields_names
+                        and 'employee' in Transaction().context):
+                    employees = values.get('employees')
+                    if not employees:
+                        employees = cls.read([user_id],
+                            ['employees'])[0]['employees']
+                    employee_id = Transaction().context['employee']
+                    if employee_id and employee_id in employees:
+                        values['employee'] = employee_id
         return result
 
 
@@ -239,7 +260,7 @@ class Property:
     company = fields.Many2One('company.company', 'Company',
         domain=[
             ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', 0)),
+                Eval('context', {}).get('company', -1)),
             ])
 
     @classmethod
@@ -256,11 +277,11 @@ class Property:
 
     @classmethod
     def search(cls, domain, offset=0, limit=None, order=None, count=False,
-            query_string=False):
+            query=False):
         if Transaction().user == 0 and not 'user' in Transaction().context:
             domain = ['AND', domain[:], ('company', '=', None)]
         return super(Property, cls).search(domain, offset=offset, limit=limit,
-            order=order, count=count, query_string=query_string)
+            order=order, count=count, query=query)
 
 
 class Sequence:
@@ -268,7 +289,7 @@ class Sequence:
     company = fields.Many2One('company.company', 'Company',
         domain=[
             ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', 0)),
+                Eval('context', {}).get('company', -1)),
             ])
 
     @classmethod
@@ -285,6 +306,21 @@ class SequenceStrict(Sequence):
     __name__ = 'ir.sequence.strict'
 
 
+class Date:
+    __name__ = 'ir.date'
+
+    @classmethod
+    def today(cls, timezone=None):
+        pool = Pool()
+        Company = pool.get('company.company')
+        company_id = Transaction().context.get('company')
+        if timezone is None and company_id:
+            company = Company(company_id)
+            if company.timezone:
+                timezone = pytz.timezone(company.timezone)
+        return super(Date, cls).today(timezone=timezone)
+
+
 class CompanyConfigStart(ModelView):
     'Company Config'
     __name__ = 'company.company.config.start'
